import dataclasses
import datetime as dt
import os
import pathlib
import unittest
from typing import TYPE_CHECKING
from unittest.mock import patch

from returns.result import Failure, ResultE, Success

from ...entities import NWPDimensionCoordinateMap, Parameter
from .ecmwf_realtime import ECMWFRealTimeS3RawRepository

if TYPE_CHECKING:
    import xarray as xr

    from nwp_consumer.internal import entities


class TestECMWFRealTimeS3RawRepository(unittest.TestCase):
    """Test the business methods of the ECMWFRealTimeS3RawRepository class."""

    @unittest.skipIf(
        condition="CI" in os.environ,
        reason="Skipping integration test that requires S3 access.",
    )  # TODO: Move into integration tests, or remove
    def test__download_and_convert(self) -> None:
        """Test the _download_and_convert method."""

        auth_result = ECMWFRealTimeS3RawRepository.authenticate()
        self.assertIsInstance(auth_result, Success, msg=f"{auth_result!s}")
        c: ECMWFRealTimeS3RawRepository = auth_result.unwrap()

        test_it: dt.datetime = dt.datetime(2024, 10, 25, 0, tzinfo=dt.UTC)
        test_coordinates: entities.NWPDimensionCoordinateMap = dataclasses.replace(
            c.model().expected_coordinates,
            init_time=[test_it],
        )

        urls: list[str] = [
            f"s3://{f}"
            for f in c._fs.ls(f"{c.bucket}/ecmwf")
            if c._wanted_file(
                filename=f.split("/")[-1],
                it=test_it,
                max_step=max(c.model().expected_coordinates.step),
            )
        ]

        for url in urls:
            with self.subTest(url=url):
                result = c._download_and_convert(url)

                self.assertIsInstance(result, Success, msg=f"{result!s}")

                da: xr.DataArray = result.unwrap()[0]
                determine_region_result = NWPDimensionCoordinateMap.from_xarray(da).bind(
                    test_coordinates.determine_region,
                )
                self.assertIsInstance(
                    determine_region_result,
                    Success,
                    msg=f"{determine_region_result!s}",
                )

    def test__wanted_file(self) -> None:
        """Test the _wanted_file method."""

        @dataclasses.dataclass
        class TestCase:
            name: str
            filename: str
            expected: bool

        test_it: dt.datetime = dt.datetime(2024, 10, 25, 0, tzinfo=dt.UTC)

        tests: list[TestCase] = [
            TestCase(
                name="valid_filename",
                filename=f"A2D{test_it:%m%d%H%M}102516001",
                expected=True,
            ),
            TestCase(
                name="invalid_init_time",
                filename="A2D09250600102516002",
                expected=False,
            ),
            TestCase(
                name="invalid_prefix",
                filename=f"GGC{test_it:%m%d%H%M}102516002",
                expected=False,
            ),
            TestCase(
                name="unexpected_extension",
                filename="A2D10251200102516001.nc",
                expected=False,
            ),
            TestCase(
                name="step_too_large",
                filename="A2D10251200102916001",
                expected=False,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.name):
                result = ECMWFRealTimeS3RawRepository._wanted_file(
                    filename=t.filename,
                    it=test_it,
                    max_step=max(ECMWFRealTimeS3RawRepository.model().expected_coordinates.step),
                )
                self.assertEqual(result, t.expected)

    def test__convert(self) -> None:
        """Test the _convert method."""

        @dataclasses.dataclass
        class TestCase:
            filename: str
            expected_coords: NWPDimensionCoordinateMap
            should_error: bool

        tests: list[TestCase] = [
            TestCase(
                filename="test_ECMWFRealtime_HRES-IFS_ssrd_20241104T00_S60.grib",
                expected_coords=dataclasses.replace(
                    ECMWFRealTimeS3RawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2024, 11, 4, 0, tzinfo=dt.UTC)],
                    variable=[Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL],
                    step=[60],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_ECMWFRealtime_HRES-IFS_10u_20241104T00_S60.grib",
                expected_coords=dataclasses.replace(
                    ECMWFRealTimeS3RawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2024, 11, 4, 0, tzinfo=dt.UTC)],
                    variable=[Parameter.WIND_U_COMPONENT_10m],
                    step=[60],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_NOAAS3_HRES-GFS_10u_20210509T06_S00.grib",
                expected_coords=ECMWFRealTimeS3RawRepository.model().expected_coordinates,
                should_error=True,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.filename):
                # Attempt to convert the file
                result = ECMWFRealTimeS3RawRepository._convert(
                    path=pathlib.Path(__file__).parent.absolute() / "test_gribs" / t.filename,
                )
                region_result: ResultE[dict[str, slice]] = result.do(
                    region
                    for das in result
                    for da in das
                    for region in NWPDimensionCoordinateMap.from_xarray(da).bind(
                        t.expected_coords.determine_region,
                    )
                )
                if t.should_error:
                    self.assertIsInstance(region_result, Failure, msg=f"{region_result}")
                else:
                    self.assertIsInstance(region_result, Success, msg=f"{region_result}")

    @patch.dict(os.environ, {"MODEL": "hres-ifs-india"}, clear=True)
    def test_convert_india(self) -> None:
        return self.test__convert()



"""Implementation of adaptors for driving actors.

Driving actors
--------------

A driving actor is an external component that initiates interaction
with the core logic. Also referred to as *primary* actors, a driving
actor represents an entrypoint that uses the core driving ports
(see `nwp_consumer.internal.ports.services`) in its implementation.
In this manner, it *handles* whatever input it receives and *drives*
the core logic to perform the necessary operations, hence the module
name.

Examples of driving or primary actors include:

- a REST services receiving requests
- a CLI tool processing user input

This module
-----------

This module contains implementations for the following driving actors:

- Command-line interface (CLI) - `nwp_consumer.internal.handlers.cli`
"""

from .cli import CLIHandler

__all__ = [
    "CLIHandler"
]



"""Domain classes for Models and their metadata.

Sources of NWP data have attributes both pertaining to and apart from
the data they deliver. This module defines classes for metadata that
tracks relevant information about the model repository and the data
it provides. This might be helpful in determining the quality of the
data, defining pipelines for processing, or establishing the availability
for a live service.

In this instance, the `ModelMetadata` refers to information pertaining
to the model used to generate the data itself.
"""

import dataclasses
import datetime as dt
import logging

import numpy as np
import pandas as pd

from .coordinates import NWPDimensionCoordinateMap
from .parameters import Parameter

log = logging.getLogger("nwp-consumer")


@dataclasses.dataclass(slots=True)
class ModelMetadata:
    """Metadata for an NWP model."""

    name: str
    """The name of the model.

    Used to name the tensor in the zarr store.
    """

    resolution: str
    """The resolution of the model with units."""

    expected_coordinates: NWPDimensionCoordinateMap
    """The expected dimension coordinate mapping.

    This is a dictionary mapping dimension labels to their coordinate values,
    for a single init time dataset, e.g.

    >>> {
    >>>     "init_time": [dt.datetime(2021, 1, 1, 0, 0), ...],
    >>>     "step": [1, 2, ...],
    >>>     "latitude": [90, 89.75, 89.5, ...],
    >>>     "longitude": [180, 179, ...],
    >>> }

    To work this out, it can be useful to use the 'grib_ls' tool from eccodes:

    >>> grib_ls -n geography -wcount=13 raw_file.grib

    Which prints grid data from the grib file.
    """

    running_hours: list[int]
    """The hours of the day that the model runs.

    Raw Repositories that provide data for the model may not have every running time.
    In this instance, use `with_running_hours` to specify the running hours specific
    to the repository.
    """

    delay_minutes: int
    """The model delay in minutes.

    Note that this can be overwritten by using `with_delay_minutes`.
    """

    chunk_count_overrides: dict[str, int] = dataclasses.field(default_factory=dict)
    """Mapping of dimension names to the desired number of chunks in that dimension.

    Overrides the default chunking strategy.

    See Also:
        - `entities.coordinates.NWPDimensionCoordinateMap.chunking`
    """

    def __str__(self) -> str:
        """Return a pretty-printed string representation of the metadata."""
        pretty: str = "".join(
            (
                "Model:",
                "\n\t{self.name} ({self.resolution} resolution)",
                "\tCoordinates:",
                "\n".join(
                    f"\t\t{dim}: {vals}"
                    if len(vals) < 5
                    else f"\t\t{dim}: {vals[:3]} ... {vals[-3:]}"
                    for dim, vals in self.expected_coordinates.__dict__.items()
                ),
            ),
        )
        return pretty

    def with_region(self, region: str) -> "ModelMetadata":
        """Returns metadata for the given model cropped to the given region.

        If an unknown region is given, the metadata is returned unchanged,
        with a warning.
        """
        match region:
            case "uk":
                return (
                    self.expected_coordinates.crop(
                        north=62,
                        west=-12,
                        south=48,
                        east=3,
                    )
                    .map(
                        lambda coords: dataclasses.replace(
                            self,
                            name=f"{self.name}_uk",
                            expected_coordinates=coords,
                        ),
                    )
                    .unwrap()
                )
            case "uk-north60":
                # same as uk. but north is 60, not 62
                return (
                    self.expected_coordinates.crop(
                        north=60,
                        west=-12,
                        south=48,
                        east=3,
                    )
                    .map(
                        lambda coords: dataclasses.replace(
                            self,
                            name=f"{self.name}_uk",
                            expected_coordinates=coords,
                        ),
                    )
                    .unwrap()
                )
            case "india":
                return (
                    self.expected_coordinates.crop(
                        north=35,
                        west=67,
                        south=6,
                        east=97,
                    )
                    .map(
                        lambda coords: dataclasses.replace(
                            self,
                            name=f"{self.name}_india",
                            expected_coordinates=coords,
                        ),
                    )
                    .unwrap()
                )
            case "west-europe":
                return (
                    self.expected_coordinates.crop(
                        north=63,
                        west=-12,
                        south=35,
                        east=26,
                    )
                    .map(
                        lambda coords: dataclasses.replace(
                            self,
                            name=f"{self.name}_west-europe",
                            expected_coordinates=coords,
                        ),
                    )
                    .unwrap()
                )
            case "nl":
                return (
                    self.expected_coordinates.crop(
                        north=54,
                        west=2,
                        south=50,
                        east=8,
                    )
                    .map(
                        lambda coords: dataclasses.replace(
                            self,
                            name=f"{self.name}_nl",
                            expected_coordinates=coords,
                        ),
                    )
                    .unwrap()
                )
            case _:
                log.warning(f"Unknown region '{region}', not cropping expected coordinates.")
                return self

    def with_chunk_count_overrides(self, overrides: dict[str, int]) -> "ModelMetadata":
        """Returns metadata for the given model with the given chunk count overrides."""
        if not set(overrides.keys()).issubset(self.expected_coordinates.dims):
            log.warning(
                "Chunk count overrides contain keys not in the expected coordinates. "
                "These will not modify the chunking strategy.",
            )
        return dataclasses.replace(self, chunk_count_overrides=overrides)

    def with_running_hours(self, hours: list[int]) -> "ModelMetadata":
        """Returns metadata for the given model with the given running hours."""
        return dataclasses.replace(self, running_hours=hours)

    def with_delay_minutes(self, delay_minutes: int) -> "ModelMetadata":
        """Returns metadata for the given model with the given delay minutes."""
        return dataclasses.replace(self, delay_minutes=delay_minutes)

    def with_max_step(self, max_step: int) -> "ModelMetadata":
        """Returns metadata for the given model with the given max step."""
        return dataclasses.replace(
            self,
            expected_coordinates=dataclasses.replace(
                self.expected_coordinates,
                step=[s for s in self.expected_coordinates.step if s <= max_step],
            ),
        )

    def month_its(self, year: int, month: int) -> list[dt.datetime]:
        """Generate all init times for a given month."""
        days = pd.Period(f"{year}-{month}").days_in_month
        its: list[dt.datetime] = []
        for day in range(1, days + 1):
            for hour in self.running_hours:
                its.append(dt.datetime(year, month, day, hour, tzinfo=dt.UTC))
        return its


class Models:
    """Namespace containing known models."""

    ECMWF_HRES_IFS_0P1DEGREE: ModelMetadata = ModelMetadata(
        name="hres-ifs",
        resolution="0.1 degrees",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 85, 1)),
            variable=[
                Parameter.WIND_U_COMPONENT_10m,
                Parameter.WIND_V_COMPONENT_10m,
                Parameter.WIND_U_COMPONENT_100m,
                Parameter.WIND_V_COMPONENT_100m,
                Parameter.WIND_U_COMPONENT_200m,
                Parameter.WIND_V_COMPONENT_200m,
                Parameter.TEMPERATURE_SL,
                Parameter.TOTAL_PRECIPITATION_RATE_GL,
                Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                Parameter.DOWNWARD_LONGWAVE_RADIATION_FLUX_GL,
                Parameter.CLOUD_COVER_HIGH,
                Parameter.CLOUD_COVER_MEDIUM,
                Parameter.CLOUD_COVER_LOW,
                Parameter.CLOUD_COVER_TOTAL,
                Parameter.SNOW_DEPTH_GL,
                Parameter.VISIBILITY_SL,
                Parameter.DIRECT_SHORTWAVE_RADIATION_FLUX_GL,
                Parameter.DOWNWARD_ULTRAVIOLET_RADIATION_FLUX_GL,
            ],
            latitude=[float(f"{lat / 10:.2f}") for lat in range(900, -900 - 1, -1)],
            longitude=[float(f"{lon / 10:.2f}") for lon in range(-1800, 1800 + 1, 1)],
        ),
        running_hours=[0, 6, 12, 18],
        delay_minutes=(60 * 26),  # 1 day, plus leeway
    )
    """ECMWF's High Resolution Integrated Forecast System."""

    ECMWF_ENS_STAT_0P1DEGREE: ModelMetadata = ModelMetadata(
        name="ens-stat",
        resolution="0.1 degrees",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 85, 3)),
            variable=[
                Parameter.PRESSURE_MSL,
                Parameter.WIND_SPEED_10m,
                Parameter.WIND_SPEED_100m,
                Parameter.TEMPERATURE_SL,
            ],
            ensemble_stat=["mean", "std", "P10", "P25", "P75", "P90"],
            latitude=[v / 10 for v in range(900, -900, -1)],
            longitude=[v / 10 for v in range(-1800, 1800, 1)],
        ),
        running_hours=[0, 12],
        delay_minutes=(60 * 26),  # 1 day, plus leeway
    )
    """Summary statistics from ECMWF's Ensemble Forecast System."""

    ECMWF_ENS_0P1DEGREE: ModelMetadata = ModelMetadata(
        name="ens",
        resolution="0.1 degrees",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 56, 1)),
            variable=[
                Parameter.DOWNWARD_LONGWAVE_RADIATION_FLUX_GL,
                Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                Parameter.DOWNWARD_ULTRAVIOLET_RADIATION_FLUX_GL,
                Parameter.DIRECT_SHORTWAVE_RADIATION_FLUX_GL,
                # Parameter.WIND_U_COMPONENT_10m,
                # Parameter.WIND_V_COMPONENT_10m,
                # Parameter.SNOW_DEPTH_GL,
                # Parameter.CLOUD_COVER_HIGH,
                # Parameter.CLOUD_COVER_MEDIUM,
                Parameter.CLOUD_COVER_LOW,
                # Parameter.CLOUD_COVER_TOTAL,
                Parameter.TEMPERATURE_SL,
                # Parameter.TOTAL_PRECIPITATION_RATE_GL,
            ],
            ensemble_member=list(range(1, 51)),
            latitude=[v / 10 for v in range(900, -900, -1)],
            longitude=[v / 10 for v in range(-1800, 1800, 1)],
        ),
        running_hours=[0, 6, 12, 18],
        delay_minutes=(60 * 26),  # 1 day, plus leeway
    )
    """Full ensemble data from ECMWF's Ensemble Forecast System."""

    NCEP_GFS_1DEGREE: ModelMetadata = ModelMetadata(
        name="ncep-gfs",
        resolution="1 degree",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(3, 49, 3)),
            variable=sorted(
                [
                    Parameter.TEMPERATURE_SL,
                    Parameter.CLOUD_COVER_TOTAL,
                    Parameter.CLOUD_COVER_HIGH,
                    Parameter.CLOUD_COVER_MEDIUM,
                    Parameter.CLOUD_COVER_LOW,
                    Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                    Parameter.DOWNWARD_LONGWAVE_RADIATION_FLUX_GL,
                    Parameter.TOTAL_PRECIPITATION_RATE_GL,
                    Parameter.SNOW_DEPTH_GL,
                    Parameter.RELATIVE_HUMIDITY_SL,
                    Parameter.VISIBILITY_SL,
                    Parameter.WIND_U_COMPONENT_10m,
                    Parameter.WIND_V_COMPONENT_10m,
                    Parameter.WIND_U_COMPONENT_100m,
                    Parameter.WIND_V_COMPONENT_100m,
                ],
            ),
            latitude=[float(lat) for lat in range(90, -90 - 1, -1)],
            longitude=[float(lon) for lon in range(-180, 180 + 1, 1)],
        ),
        running_hours=[0, 6, 12, 18],
        delay_minutes=(60 * 5),  # 5 hours
    )
    """NCEP's Global Forecast System."""

    MO_UM_GLOBAL_17KM: ModelMetadata = ModelMetadata(
        name="um-global",
        resolution="17km",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 48, 1)),
            variable=[
                Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                Parameter.CLOUD_COVER_TOTAL,
                Parameter.CLOUD_COVER_HIGH,
                Parameter.CLOUD_COVER_LOW,
                Parameter.CLOUD_COVER_MEDIUM,
                Parameter.RELATIVE_HUMIDITY_SL,
                Parameter.SNOW_DEPTH_GL,
                Parameter.TEMPERATURE_SL,
                Parameter.WIND_U_COMPONENT_10m,
                Parameter.WIND_V_COMPONENT_10m,
                Parameter.VISIBILITY_SL,
            ],
            latitude=[float(f"{lat:.4f}") for lat in np.arange(89.856, -89.856 - 0.156, -0.156)],
            longitude=[
                float(f"{lon:.4f}")
                for lon in np.concatenate(
                    [
                        np.arange(-45, 45, 0.234),
                        np.arange(45, 135, 0.234),
                        np.arange(135, 225, 0.234),
                        np.arange(225, 315, 0.234),
                    ],
                )
            ],
            # TODO: Change to -180 -> 180
        ),
        running_hours=[0, 6, 12, 18],
        delay_minutes=300,
    )
    """MetOffice's Unified Model, in the Global configuration, at a resolution of 17km."""

    MO_UM_GLOBAL_10KM: ModelMetadata = ModelMetadata(
        name="um-global",
        resolution="10km",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 55)),
            variable=sorted(
                [
                    Parameter.CLOUD_COVER_TOTAL,
                    Parameter.CLOUD_COVER_HIGH,
                    Parameter.CLOUD_COVER_MEDIUM,
                    Parameter.CLOUD_COVER_LOW,
                    Parameter.VISIBILITY_SL,
                    Parameter.RELATIVE_HUMIDITY_SL,
                    Parameter.SNOW_DEPTH_GL,
                    Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                    Parameter.TEMPERATURE_SL,
                    Parameter.WIND_U_COMPONENT_10m,
                    Parameter.WIND_V_COMPONENT_10m,
                ],
            ),
            latitude=[
                float(f"{lat:.4f}") for lat in np.arange(89.953125, -89.953125 - 0.09375, -0.09375)
            ],
            longitude=[
                float(f"{lon:.4f}")
                for lon in np.arange(-179.929687, 179.929688 + 0.140625, 0.140625)
            ],
        ),
        running_hours=[0, 12],
        delay_minutes=300,
    )
    """MetOffice's Unified Model, in the Global configuration, at a resolution of 10km."""

    MO_UM_UKV_2KM_OSGB: ModelMetadata = ModelMetadata(
        name="um-ukv",
        resolution="2km",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 49)),
            variable=sorted(
                [
                    Parameter.CLOUD_COVER_TOTAL,
                    Parameter.CLOUD_COVER_HIGH,
                    Parameter.CLOUD_COVER_MEDIUM,
                    Parameter.CLOUD_COVER_LOW,
                    Parameter.VISIBILITY_SL,
                    Parameter.RELATIVE_HUMIDITY_SL,
                    Parameter.SNOW_DEPTH_GL,
                    Parameter.DOWNWARD_LONGWAVE_RADIATION_FLUX_GL,
                    Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                    Parameter.TEMPERATURE_SL,
                    Parameter.WIND_U_COMPONENT_10m,
                    Parameter.WIND_V_COMPONENT_10m,
                    Parameter.WIND_DIRECTION_10m,
                    Parameter.WIND_SPEED_10m,
                    Parameter.TOTAL_PRECIPITATION_RATE_GL,
                ],
            ),
            # Taken from page 4 of https://zenodo.org/record/7357056
            y_osgb=[int(y) for y in np.arange(start=1223000, stop=-185000, step=-2000)],
            x_osgb=[int(x) for x in np.arange(start=-239000, stop=857000, step=2000)],
        ),
        running_hours=list(range(0, 24, 6)),
        delay_minutes=120,
    )
    """MetOffice's Unified Model in the UKV configuration, at a resolution of 2km"""

    MO_UM_UKV_2KM_LAEA: ModelMetadata = ModelMetadata(
        name="um-ukv",
        resolution="2km",
        expected_coordinates=NWPDimensionCoordinateMap(
            init_time=[],
            step=list(range(0, 43)),
            variable=sorted(
                [
                    Parameter.CLOUD_COVER_HIGH,
                    Parameter.CLOUD_COVER_MEDIUM,
                    Parameter.CLOUD_COVER_LOW,
                    Parameter.VISIBILITY_SL,
                    Parameter.RELATIVE_HUMIDITY_SL,
                    Parameter.SNOW_DEPTH_GL,
                    Parameter.DOWNWARD_LONGWAVE_RADIATION_FLUX_GL,
                    Parameter.DOWNWARD_SHORTWAVE_RADIATION_FLUX_GL,
                    Parameter.TEMPERATURE_SL,
                    Parameter.WIND_DIRECTION_10m,
                    Parameter.WIND_SPEED_10m,
                    Parameter.TOTAL_PRECIPITATION_RATE_GL,
                ],
            ),
            # Taken from iris-grib reading in MetOffice UKV data
            y_laea=[int(y) for y in np.arange(start=700000, stop=-576000 - 2000, step=-2000)],
            x_laea=[int(x) for x in np.arange(start=-576000, stop=332000 + 2000, step=2000)],
        ),
        running_hours=list(range(0, 24, 3)),  # Only first 12 steps available for hourly runs
        delay_minutes=120,
    )
    """MetOffice's Unified Model in the UKV configuration, at a resolution of 2km"""



"""Adaptor for the CLI driving actor."""

import argparse
import datetime as dt
import logging

from returns.result import Failure, ResultE

from nwp_consumer.internal import ports, services

log = logging.getLogger("nwp-consumer")


class CLIHandler:
    """CLI driving actor."""

    model_adaptor: type[ports.RawRepository]
    notification_adaptor: type[ports.NotificationRepository]

    def __init__(
        self,
        model_adaptor: type[ports.RawRepository],
        notification_adaptor: type[ports.NotificationRepository],
    ) -> None:
        """Create a new instance."""
        self.model_adaptor = model_adaptor
        self.notification_adaptor = notification_adaptor

    @property
    def parser(self) -> argparse.ArgumentParser:
        """Return the CLI argument parser."""
        parser = argparse.ArgumentParser(description="NWP Consumer CLI")
        subparsers = parser.add_subparsers(dest="command")

        consume_command = subparsers.add_parser(
            "consume",
            help="Consume NWP data for a single init time",
        )
        consume_command.add_argument(
            "--init-time",
            "-i",
            help="Initialization time of the forecast (YYYY-MM-DDTHH). "
            "Omit to pull the latest available forecast.",
            type=dt.datetime.fromisoformat,
            required=False,
        )
        consume_command.add_argument(
            "--keep-failed",
            help="Don't delete the data if consuming fails",
            action="store_true",
        )

        archive_command = subparsers.add_parser(
            "archive",
            help="Archive NWP data for a given month",
        )
        archive_command.add_argument(
            "--year",
            "-y",
            help="Year to archive",
            type=int,
            required=True,
        )
        archive_command.add_argument(
            "--month",
            "-m",
            help="Month to archive",
            type=int,
            required=True,
        )
        archive_command.add_argument(
            "--keep-failed",
            help="Don't delete the data if archiving fails",
            action="store_true",
        )

        info_command = subparsers.add_parser("info", help="Show model repository info")
        info_options = info_command.add_mutually_exclusive_group()
        info_options.add_argument(
            "--model",
            help="Show information about the selected model repository.",
            action="store_true",
        )
        info_options.add_argument(
            "--parameters",
            help="Show information about all available parameters.",
            action="store_true",
        )

        return parser

    def run(self) -> int:
        """Run the CLI handler.

        Returns the appropriate exit code.
        """
        args = self.parser.parse_args()
        match args.command:
            case "consume":
                service_result = services.ConsumerService.from_adaptors(
                    model_adaptor=self.model_adaptor,
                    notification_adaptor=self.notification_adaptor,
                )
                result: ResultE[str] = service_result.do(
                    consume_result
                    for service in service_result
                    for consume_result in service.consume(
                        period=args.init_time,
                        delete_on_failure=not args.keep_failed,
                    )
                )
                if isinstance(result, Failure):
                    log.error(f"Failed to consume NWP data: {result!s}")
                    return 1

            case "archive":
                service_result = services.ConsumerService.from_adaptors(
                    model_adaptor=self.model_adaptor,
                    notification_adaptor=self.notification_adaptor,
                )
                period: dt.date = dt.date(args.year, args.month, 1)
                result = service_result.do(
                    consume_result
                    for service in service_result
                    for consume_result in service.consume(
                        period=period,
                        delete_on_failure=not args.keep_failed,
                    )
                )
                if isinstance(result, Failure):
                    log.error(f"Failed to archive NWP data: {result!s}")
                    return 1

            case "info":
                log.error("Info command is coming soon! :)")

            case _:
                log.error(f"Unknown command: {args.command}")
                self.parser.print_help()
                return 1

        return 0



"""Service interfaces for consumer services.

These interfaces define the signatures that *driving* actors must conform to
in order to interact with the core.

Sometimes referred to as *primary ports*.
"""

import abc
import datetime as dt

from returns.result import ResultE


class ConsumeUseCase(abc.ABC):
    """Interface for the consumer use case.

    Defines the business-critical methods for the following use cases:

    - 'A user should be able to consume NWP data for a given initialization time.'
    """

    @abc.abstractmethod
    def consume(
        self,
        period: dt.datetime | dt.date | None = None,
    ) -> ResultE[str]:
        """Consume NWP data to Zarr format for desired time period.

        Where possible the implementation should be as memory-efficient as possible.
        The designs of the repository methods also enable parallel processing within
        the implementation.

        Args:
            period: The period for which to gather init time data.

        Returns:
            The path to the produced Zarr store.

        See Also:
            - `repositories.RawRepository.fetch_init_data`
            - `tensorstore.TensorStore.write_to_region`
            - https://joblib.readthedocs.io/en/stable/auto_examples/parallel_generator.html
        """
        pass

    @abc.abstractmethod
    def archive(self, period: dt.date) -> ResultE[str]:
        """Archive NWP data to Zarr format for desired time period.

        Args:
            period: The period for which to gather init time data.

        Returns:
            The path to the produced Zarr store.
        """
        pass



"""Domain entities for service notifications.

Upon completion of the processing of a request, the service can
deliver a notification of the result to a notification repository.
This module defines the structure of the notification messages.
"""

import dataclasses


@dataclasses.dataclass
class PerformanceMetadata:
    """Metadata for a service operation."""

    duration_seconds: float
    """The duration of the operation in seconds."""

    memory_mb: float
    """The memory usage of the operation in megabytes."""


@dataclasses.dataclass(slots=True)
class StoreCreatedNotification:
    """A notification of successful store creation."""

    filename: str
    """The name of the store created, including extension."""

    size_mb: int
    """The size of the store in megabytes."""

    performance: PerformanceMetadata
    """Metadata for the operation."""

    def __str__(self) -> str:
        """Return a string representation of the notification."""
        return "".join(
            (
                f"Store created: {self.filename} ({self.size_mb} MB) in ",
                f"{self.performance.duration_seconds} secs ",
                f"(using {self.performance.memory_mb} MB RAM)",
            ),
        )


@dataclasses.dataclass(slots=True)
class StoreAppendedNotification:
    """A notification of successful append to a store."""

    filename: str
    """The name of the store appended to, including extension."""

    size_mb: int
    """The size of the store in megabytes."""

    performance: PerformanceMetadata
    """Metadata for the operation."""



"""Model repository implementation for ECMWF live data from S3.

Repository Information
======================

Documented Structure
--------------------

When getting live or realtime data from ECMWF, grib files are sent by
a data provider to a location of choice, in this case an S3 bucket.
The `ECMWF Dissemination Schedule <https://confluence.ecmwf.int/display/DAC/Dissemination+schedule>`_
describes the naming convention and time ordering for these files:

- A 2-character prefix
- A 1-character dissemination stream indicator
- 8 digits representing the initialization time in the format mmddHHMM
- 8 digits representing the target time in the format mmddHHMM
- 1 digit representing the file number(?)

So a file named ``A2D10250000D10260100`` would be for an initialization
time of 2024-10-25 00:00 and a target time of 2024-10-26 01:00 (step of 25 hours).

The file contents is specific to the order agreed with the data provider.
For the order that OCF has created, there are four distinct datasets.
This is because OCF has ordered two separate regions and 17 variables,
which are split across two datasets.

Also, some of the data contains larger steps than we are interested in due
to necessities in the order creation process.

"""

import datetime as dt
import logging
import os
import pathlib
import re
from collections.abc import Callable, Iterator
from typing import override

import cfgrib
import numpy as np
import s3fs
import xarray as xr
from joblib import delayed
from returns.result import Failure, ResultE, Success

from nwp_consumer.internal import entities, ports

log = logging.getLogger("nwp-consumer")


class ECMWFRealTimeS3RawRepository(ports.RawRepository):
    """Model repository implementation for ECMWF live data from S3."""

    bucket: str
    _fs: s3fs.S3FileSystem

    def __init__(self, bucket: str, fs: s3fs.S3FileSystem) -> None:
        """Create a new instance of the class."""
        self.bucket = bucket
        self._fs = fs

    @staticmethod
    @override
    def repository() -> entities.RawRepositoryMetadata:
        return entities.RawRepositoryMetadata(
            name="ECMWF-Realtime-S3",
            is_archive=False,
            is_order_based=True,
            max_connections=100,
            required_env=[
                "ECMWF_REALTIME_S3_ACCESS_KEY",
                "ECMWF_REALTIME_S3_ACCESS_SECRET",
                "ECMWF_REALTIME_S3_BUCKET",
                "ECMWF_REALTIME_S3_REGION",
            ],
            optional_env={
                "ECMWF_REALTIME_DISSEMINATION_FILE_PREFIX": "A2",
                "ECMWF_REALTIME_S3_BUCKET_PREFIX": "ecmwf",
            },
            postprocess_options=entities.PostProcessOptions(),
            available_models={
                "default": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region("uk-north60").
                    with_delay_minutes(60 * 7),
                "hres-ifs-uk": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region("uk-north60").
                    with_delay_minutes(60 * 7),
                "hres-ifs-india": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region(
                    "india",
                ).with_chunk_count_overrides({"variable": 1}).with_delay_minutes(60 * 7),
                "hres-ifs-nl": entities.Models.ECMWF_HRES_IFS_0P1DEGREE.with_region(
                    "nl",
                ).with_max_step(84).with_delay_minutes(60 * 7),
            },
        )

    @staticmethod
    @override
    def model() -> entities.ModelMetadata:
        requested_model: str = os.getenv("MODEL", default="default")
        if requested_model not in ECMWFRealTimeS3RawRepository.repository().available_models:
            log.warning(
                f"Unknown model '{requested_model}' requested, falling back to default ",
                "ECMWF Realtime S3 repository only supports "
                f"'{list(ECMWFRealTimeS3RawRepository.repository().available_models.keys())}'. "
                "Ensure MODEL environment variable is set to a valid model name.",
            )
            requested_model = "default"
        return ECMWFRealTimeS3RawRepository.repository().available_models[requested_model]

    @override
    def fetch_init_data(
        self, it: dt.datetime,
    ) -> Iterator[Callable[..., ResultE[list[xr.DataArray]]]]:
        # List relevant files in the S3 bucket
        try:
            urls: list[str] = [
                f"s3://{f}"
                for f in self._fs.ls(f"{self.bucket}/ecmwf")
                if self._wanted_file(
                    filename=f.split("/")[-1],
                    it=it,
                    max_step=max(self.model().expected_coordinates.step),
                )
            ]
        except Exception as e:
            yield delayed(Failure)(
                ValueError(
                    f"Failed to list files in bucket path '{self.bucket}/ecmwf'. "
                    "Ensure the path exists and the caller has relevant access permissions. "
                    f"Encountered error: {e}",
                ),
            )
            return

        if len(urls) == 0:
            yield delayed(Failure)(
                ValueError(
                    f"No raw files found for init time '{it.strftime('%Y-%m-%d %H:%M')}' "
                    f"in bucket path '{self.bucket}/ecmwf'. Ensure files exist at the given path "
                    "named with the expected pattern, e.g. 'A2S10250000102603001.",
                ),
            )
            return

        log.debug(
            f"Found {len(urls)} file(s) for init time '{it.strftime('%Y-%m-%d %H:%M')}' "
            f"in bucket path '{self.bucket}/ecmwf'.",
        )
        for url in urls:
            yield delayed(self._download_and_convert)(url=url)

    @classmethod
    @override
    def authenticate(cls) -> ResultE["ECMWFRealTimeS3RawRepository"]:
        missing_envs = cls.repository().missing_required_envs()
        if len(missing_envs) > 0:
            return Failure(
                OSError(
                    f"Cannot authenticate with ECMWF Realtime S3 service due to "
                    f"missing required environment variables: {', '.join(missing_envs)}",
                ),
            )
        try:
            bucket: str = os.environ["ECMWF_REALTIME_S3_BUCKET"]
            _fs: s3fs.S3FileSystem = s3fs.S3FileSystem(
                key=os.environ["ECMWF_REALTIME_S3_ACCESS_KEY"],
                secret=os.environ["ECMWF_REALTIME_S3_ACCESS_SECRET"],
                client_kwargs={
                    "endpoint_url": os.environ.get("AWS_ENDPOINT_URL", None),
                    "region_name": os.environ["ECMWF_REALTIME_S3_REGION"],
                },
            )
        except Exception as e:
            return Failure(
                ConnectionError(
                    "Failed to connect to S3 for ECMWF data. "
                    f"Credentials may be wrong or undefined. Encountered error: {e}",
                ),
            )

        log.debug(f"Successfully authenticated with S3 instance '{bucket}'")
        return Success(cls(bucket=bucket, fs=_fs))

    def _download_and_convert(self, url: str) -> ResultE[list[xr.DataArray]]:
        """Download and convert a file to xarray DataArrays.

        Args:
            url: The URL of the file to download.
        """
        return self._download(url=url).bind(self._convert)

    def _download(self, url: str) -> ResultE[pathlib.Path]:
        """Download an ECMWF realtime file from S3.

        Args:
            url: The URL to the S3 object.
        """
        local_path: pathlib.Path = (
            (
                pathlib.Path(
                    os.getenv(
                        "RAWDIR",
                        f"~/.local/cache/nwp/{self.repository().name}/{self.model().name}/raw",
                    ),
                )
                / url.split("/")[-1]
            )
            .with_suffix(".grib")
            .expanduser()
        )

        # Only download the file if not already present
        if local_path.exists() and local_path.stat().st_size > 0:
            log.debug("Skipping download for existing file at '%s'.", local_path.as_posix())
        else:
            local_path.parent.mkdir(parents=True, exist_ok=True)
            log.debug("Requesting file from S3 at: '%s'", url)

            try:
                if not self._fs.exists(url):
                    raise FileNotFoundError(f"File not found at '{url}'")

                log.debug("Writing file from '%s' to '%s'", url, local_path.as_posix())
                with local_path.open("wb") as lf, self._fs.open(url, "rb") as rf:
                    for chunk in iter(lambda: rf.read(12 * 1024), b""):
                        lf.write(chunk)
                        lf.flush()

            except Exception as e:
                return Failure(
                    OSError(
                        f"Failed to download file from S3 at '{url}'. Encountered error: {e}",
                    ),
                )

            if local_path.stat().st_size != self._fs.info(url)["size"]:
                return Failure(
                    ValueError(
                        f"Failed to download file from S3 at '{url}'. "
                        "File size mismatch. File may be corrupted.",
                    ),
                )

        return Success(local_path)

    @staticmethod
    def _convert(path: pathlib.Path) -> ResultE[list[xr.DataArray]]:
        """Convert a grib file to an xarray DataArray.

        Args:
            path: The path to the grib file.
        """
        try:
            dss: list[xr.Dataset] = cfgrib.open_datasets(path.as_posix())
        except Exception as e:
            return Failure(
                OSError(
                    f"Error opening '{path}' as list of xarray Datasets: {e}",
                ),
            )
        if len(dss) == 0:
            return Failure(
                ValueError(
                    f"No datasets found in '{path}'. File may be corrupted. "
                    "A redownload of the file may be required.",
                ),
            )

        processed_das: list[xr.DataArray] = []
        num_skipped: int = 0
        expected_lons = ECMWFRealTimeS3RawRepository.model().expected_coordinates.longitude
        expected_lats = ECMWFRealTimeS3RawRepository.model().expected_coordinates.latitude
        expected_steps = ECMWFRealTimeS3RawRepository.model().expected_coordinates.step

        for i, ds in enumerate(dss):
            # ECMWF Realtime provides all regions in one set of datasets,
            # so distinguish via their coordinates
            step = np.timedelta64(ds.coords["step"].values, "h").astype(int) # type: ignore[arg-type]
            is_relevant_dataset_predicate: bool = (
                (expected_lons is not None and expected_lats is not None)
                and (expected_lons[0] <= max(ds.coords["longitude"].values) <= expected_lons[-1])
                and (expected_lats[-1] <= max(ds.coords["latitude"].values) <= expected_lats[0])
                and (expected_steps[0] <= step <= expected_steps[-1])
            )
            if not is_relevant_dataset_predicate:
                num_skipped += 1
                continue
            try:
                da: xr.DataArray = (
                    entities.Parameter.rename_else_drop_ds_vars(
                        ds=ds,
                        allowed_parameters=ECMWFRealTimeS3RawRepository.model().expected_coordinates.variable,
                    )
                    .rename(name_dict={"time": "init_time"})
                    .expand_dims(dim="init_time")
                    .expand_dims(dim="step")
                    .to_dataarray(name=ECMWFRealTimeS3RawRepository.model().name)
                )
                da = (
                    da.drop_vars(
                        names=[
                            c
                            for c in ds.coords
                            if c
                            not in ECMWFRealTimeS3RawRepository.model().expected_coordinates.dims
                        ],
                        errors="ignore",
                    )
                    .transpose(*ECMWFRealTimeS3RawRepository.model().expected_coordinates.dims)
                    .sortby(variables=["step", "variable", "longitude"])
                    .sortby(variables="latitude", ascending=False)
                )

            except Exception as e:
                return Failure(
                    ValueError(
                        f"Error processing dataset {i} from '{path}' to DataArray: {e}",
                    ),
                )
            # Put each variable into its own DataArray:
            # * Each raw file does not contain a full set of parameters
            # * and so may not produce a contiguous subset of the expected coordinates.
            processed_das.extend(
                [da.where(cond=da["variable"] == v, drop=True) for v in da["variable"].values],
            )

        if len(processed_das) == 0:
            return Failure(
                ValueError(
                    f"Skipped {num_skipped}/{len(dss)} datasets from '{path}'. "
                    "File may not contain the expected parameters and geographic bounds.",
                ),
            )

        return Success(processed_das)

    @staticmethod
    def _wanted_file(filename: str, it: dt.datetime, max_step: int) -> bool:
        """Determine if the file is wanted based on the init time.

        See module docstring for the file naming convention.
        Returns True if the filename describes data corresponding to the input
        initialization time and model metadata.

        Args:
            filename: The name of the file.
            it: The init time of the model run.
            max_step: The maximum step in hours to consider.
        """
        prefix: str = os.getenv("ECMWF_REALTIME_DISSEMINATION_FILE_PREFIX", "A2")
        pattern: str = r"^" + prefix + r"[DS](\d{8})(\d{8})\d$"
        match: re.Match[str] | None = re.search(pattern=pattern, string=filename)
        if match is None:
            return False
        if it.strftime("%m%d%H%M") != match.group(1):
            return False
        tt: dt.datetime = dt.datetime.strptime(
            str(it.year) + match.group(2) + "+0000",
            "%Y%m%d%H%M%z",
        )
        return tt <= it + dt.timedelta(hours=max_step)



import dataclasses
import datetime as dt
import os
import pathlib
import unittest
from unittest.mock import patch

from returns.result import Failure, ResultE, Success

from ...entities import NWPDimensionCoordinateMap
from .ecmwf_mars import ECMWFMARSRawRepository


class TestECMWFMARSEModelREpository(unittest.TestCase):
    """Test the business methods of the ECMWFRealTimeS3RawRepository class."""

    @patch.dict(os.environ, {"MODEL": "ens-stat-uk"}, clear=True)
    def test__convert(self) -> None:
        """Test the _convert method."""

        @dataclasses.dataclass
        class TestCase:
            filename: str
            expected_coords: NWPDimensionCoordinateMap
            should_error: bool

        tests: list[TestCase] = [
            TestCase(
                filename="test_ECMWFMARS_enfo-em_t2m-si10-si100-msp_20240101T00_S03-06.grib",
                expected_coords=dataclasses.replace(
                    ECMWFMARSRawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2024, 1, 1, 0, tzinfo=dt.UTC)],
                    ensemble_stat=["mean"],
                    step=[3, 6],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_ECMWFMARS_enfo-es_t2m-si10-si100-msp_20240101T00_S03-06.grib",
                expected_coords=dataclasses.replace(
                    ECMWFMARSRawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2024, 1, 1, 0, tzinfo=dt.UTC)],
                    step=[3, 6],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_NOAAS3_HRES-GFS_10u_20210509T06_S00.grib",
                expected_coords=ECMWFMARSRawRepository.model().expected_coordinates,
                should_error=True,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.filename):
                result = ECMWFMARSRawRepository._convert(
                    path=pathlib.Path(__file__).parent.absolute() / "test_gribs" / t.filename,
                )
                region_result: ResultE[dict[str, slice]] = result.do(
                    region
                    for das in result
                    for da in das
                    for region in NWPDimensionCoordinateMap.from_xarray(da).bind(
                        t.expected_coords.determine_region,
                    )
                )
                if t.should_error:
                    self.assertIsInstance(region_result, Failure, msg=f"{region_result}")
                else:
                    self.assertIsInstance(region_result, Success, msg=f"{region_result}")



"""Internal workings of the services.

Why have an internal package?
-----------------------------

This package is meant to be run as a services, either via a binary or a container image.
However, the code can still be used as a library, and a user could import the modules
from this package and use them in their own code.

The "internal" package signifies that the modules within are not meant to be used,
or are not guaranteed to be stable, for external users. This helps to discourage casual
dependence in other services. Any functionality looking to be re-used should either
become a shared library or simply be copied from the source code.
"""

from . import entities, handlers, ports, repositories, services

__all__ = ["entities", "ports", "handlers", "repositories", "services"]



"""Domain classes for store metadata.

Converted data is stored in Zarr stores, which are chunked datastores
enabling subselection across any dimension of data, provided it is
chunked appropriately.

This module provides a class for storing metadata about a Zarr store.

TODO: 2024-11-20 This module wants refactoring into smaller testable components.
"""

import abc
import dataclasses
import datetime as dt
import logging
import os
import pathlib
import shutil
from collections.abc import Mapping, MutableMapping
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr
import zarr
from returns.result import Failure, ResultE, Success

from .coordinates import NWPDimensionCoordinateMap
from .parameters import Parameter
from .postprocess import PostProcessOptions

log = logging.getLogger("nwp-consumer")


@dataclasses.dataclass(slots=True)
class ParameterScanResult:
    """Container for the results of a scan of a parameter's values."""

    mean: float
    """The mean value of the parameter's data."""
    is_valid: bool
    """Whether the parameter's data values are valid.

    This is determined according to the parameter's limits and threshold.
    See `entities.parameters.Parameter`.
    """
    has_nulls: bool
    """Whether the parameter's data contains null values."""


@dataclasses.dataclass(slots=True)
class TensorStore(abc.ABC):
    """Store class for multidimensional data.

    This class is used to store data in a Zarr store.
    Each store instance has defined coordinates for the data,
    and is capable of handling parallel, region-based updates.
    """

    name: str
    """Identifier for the store and the data within."""

    path: str
    """The path to the store."""

    coordinate_map: NWPDimensionCoordinateMap
    """The coordinates of the store."""

    size_kb: int
    """The size of the store in kilobytes."""

    encoding: dict[str, Any]
    """The encoding passed to Zarr whilst writing."""

    @classmethod
    def initialize_empty_store(
        cls,
        model: str,
        repository: str,
        coords: NWPDimensionCoordinateMap,
        chunks: dict[str, int],
    ) -> ResultE["TensorStore"]:
        """Initialize a store for a given init time.

        This method writes a blank dataarray to disk based on the input coordinates,
        which define the dimension labels and tick values of the output dataset object.

        .. note: If a store already exists at the expected path,
           it is checked for consistency with the input coordinates and used if valid.

        The dataarray is 'blank' because it is written via::

            dataarray.to_zarr("<example>.zarr", compute=False)

        which writes the metadata alone.
        The utility of this is to enable region-based writing of new data to the store,
        which further allows for parallel write processes.

        There is a gotcha: regional writes can never be done in parallel to the same chunk,
        so writes must always be done at the chunk level or higher (as a chunk is an
        individual file in the store). To this effect, chunks are chosen to cover as small
        a unit of data as could reasonably be expected to be provided by an NWP source:

        - Raw data files may not contain the full grid of data, hence a chunk of size equal
          to a quarter the length of the grid dimension (lat/lon/x/y axes) is used.
        - Raw data files may contain as little as one step for a single parameter, so a chunk
          size of 1 is used along the step dimension.
        - As above for the init_time dimension.

        Args:
            model: The name of the model providing the tensor data.
                   This is also used as the name of the tensor.
            repository: The name of the repository providing the tensor data.
            coords: The coordinates of the store.
            chunks: The chunk sizes for the store.

        Returns:
            An indicator of a successful store write containing the number of bytes written.

        See Also:
            - https://docs.xarray.dev/en/stable/user-guide/io.html#appending-to-existing-zarr-stores
            - https://docs.xarray.dev/en/stable/user-guide/io.html#distributed-writes

        Returns:
            A new instance of the TensorStore class.
        """
        if not isinstance(coords.init_time, list) or len(coords.init_time) == 0:
            return Failure(
                ValueError(
                    "Cannot initialize store with 'init_time' dimension coordinates not "
                    "specified via a populated list. Check instantiation of "
                    "NWPDimensionCoordinateMap passed to this function. "
                    f"Got: {coords.init_time} (not a list, or empty).",
                ),
            )

        zarrdir = os.getenv("ZARRDIR", f"~/.local/cache/nwp/{repository}/{model}/data")
        store: zarr.storage.Store
        path: str
        filename: str = TensorStore.gen_store_filename(coords=coords)
        try:
            if zarrdir.startswith("s3"):
                store_result = cls._create_zarrstore_s3(zarrdir, filename)
                store, path = store_result.unwrap()  # Can do this as exceptions are caught
            else:
                path = pathlib.Path("/".join((zarrdir, filename))).expanduser().as_posix()
                store = zarr.storage.DirectoryStore(path)
        except Exception as e:
            return Failure(
                OSError(
                    f"Unable to create Directory Store at dir '{zarrdir}'. "
                    "Ensure ZARRDIR environment variable is specified correctly. "
                    f"Error context: {e}",
                ),
            )

        # Write the coordinates to a skeleton Zarr store
        # * 'compute=False' enables only saving metadata
        # * 'mode="w-"' fails if it finds an existing store

        da: xr.DataArray = coords.as_zeroed_dataarray(name=model, chunks=chunks)
        encoding = {
            model: {"write_empty_chunks": False},
            "init_time": {"units": "nanoseconds since 1970-01-01"},
            "step": {"units": "hours"},
        }
        try:
            _ = da.to_zarr(
                store=store,
                compute=False,
                mode="w-",
                consolidated=True,
                encoding=encoding,
            )
            log.info("Created blank zarr store at '%s'", path)
            # Ensure the store is readable
            store_da = xr.open_dataarray(store, engine="zarr")
        except zarr.errors.ContainsGroupError:
            store_da = xr.open_dataarray(store, engine="zarr")
            if store_da.name != da.name:  # TODO: Also check for equality of coordinates
                return Failure(
                    OSError(
                        f"Existing store at '{path}' is for a different model. "
                        "Delete the existing store or move it to a new location, "
                        "or choose a new location for the new store via ZARRDIR.",
                    ),
                )
            log.info(f"Using existing store at '{path}'")
            return Success(
                cls(
                    name=model,
                    path=path,
                    coordinate_map=coords,
                    size_kb=store_da.nbytes // 1024,
                    encoding=encoding,
                ),
            )
        except Exception as e:
            return Failure(
                OSError(
                    f"Failed writing blank store to '{path}': {e}",
                ),
            )

        # Check the resultant array's coordinates can be converted back
        coordinate_map_result = NWPDimensionCoordinateMap.from_xarray(store_da)
        if isinstance(coordinate_map_result, Failure):
            return Failure(
                OSError(
                    f"Error reading back coordinates of initialized store "
                    f"from path '{path}' (possible corruption): {coordinate_map_result}",
                ),
            )

        return Success(
            cls(
                name=model,
                path=path,
                coordinate_map=coordinate_map_result.unwrap(),
                size_kb=0,
                encoding=encoding,
            ),
        )

    # def from_existing_store(
    #    model: str,
    #    repository: str,
    #    expected_coords: NWPDimensionCoordinateMap,
    # ) -> ResultE["TensorStore"]:
    #    """Create a TensorStore instance from an existing store."""
    #    pass # TODO

    #    for dim in store_da.dims:
    #        if dim not in da.dims:
    #            return Failure(
    #                ValueError(
    #                    "Cannot use existing store due to mismatched coordinates. "
    #                    f"Dimension '{dim}' in existing store not found in new store. "
    #                    "Use 'overwrite_existing=True' or move the existing store at "
    #                    f"'{store}' to a new location. ",
    #                ),
    #            )
    #        if not np.array_equal(store_da.coords[dim].values, da.coords[dim].values):
    #            return Failure(
    #                ValueError(
    #                    "Cannot use existing store due to mismatched coordinates. "
    #                    f"Dimension '{dim}' in existing store has different coordinate "
    #                    "values from specified. "
    #                    "Use 'overwrite_existing=True' or move the existing store at "
    #                    f"'{store}' to a new location.",
    #                ),
    #            )

    # --- Business logic methods --- #
    def write_to_region(
        self,
        da: xr.DataArray,
        region: dict[str, slice] | None = None,
    ) -> ResultE[int]:
        """Write partial data to the store.

        The optional region is a dictionary which maps dimension labels to slices.
        These define the region in the store to write to.

        If the region dict is empty or not provided, the region is determined
        via the 'determine_region' method.

        This function should be thread safe, so a check is performed on the region
        to ensure that it can be safely written to in parallel, i.e. that it covers
        an integer number of chunks.

        Args:
            da: The data to write to the store.
            region: The region to write to.

        Returns:
            An indicator of a successful store write containing the number of bytes written.

        See Also:
            - https://docs.xarray.dev/en/stable/user-guide/io.html#distributed-writes
        """
        # Attempt to determine the region if missing
        if region is None or region == {}:
            region_result = NWPDimensionCoordinateMap.from_xarray(da).bind(
                self.coordinate_map.determine_region,
            )
            if isinstance(region_result, Failure):
                return region_result
            region = region_result.unwrap()

        # For each dimensional slice defining the region, check the slice represents an
        # integer number of chunks along that dimension.
        # * This is to ensure that the data can safely be written in parallel.
        # * The start and and of each slice should be divisible by the chunk size.
        chunksizes: Mapping[Any, tuple[int, ...]] = xr.open_dataarray(
            self.path,
            engine="zarr",
        ).chunksizes
        for dim, slc in region.items():
            chunk_size = chunksizes.get(dim, (1,))[0]
            # TODO: Determine if this should return a full failure object
            if slc.start % chunk_size != 0 or slc.stop % chunk_size != 0:
                log.warning(
                    f"Determined region of raw data to be written for dimension '{dim}'"
                    f"does not align with chunk boundaries of the store. "
                    f"Dimension '{dim}' has a chunk size of {chunk_size}, "
                    "but the data to be written for this dimension starts at chunk "
                    f"{slc.start / chunk_size:.2f} (index {slc.start}) and ends at chunk "
                    f"{slc.stop / chunk_size:.2f} (index {slc.stop}). "
                    "As such, this region cannot be safely written in parallel. "
                    "Ensure the chunking is granular enough to cover the raw data region.",
                )

        # Perform the regional write
        try:
            da.to_zarr(store=self.path, region=region, consolidated=True)
        except Exception as e:
            return Failure(
                OSError(
                    f"Error writing to region of store: {e}",
                ),
            )

        # Calculate the number of bytes written
        nbytes: int = da.nbytes
        del da
        self.size_kb += nbytes // 1024
        return Success(nbytes)

    @staticmethod
    def _has_nans(store_da: xr.DataArray) -> ResultE[bool]:
        """Check the store for NaN values."""
        # I don't like having environ calls buried like this - needs refactoring to the top
        nans_in_image_threshold: float = float(os.getenv("ALLOWED_NAN_PERCENTAGE", "0.05"))
        images_failing_nan_check_threshold: float = float(
            os.getenv("ALLOWED_VALIDATION_FAILURE_PERCENTAGE", "0.02"),
        )

        def _calc_null_percentage(data: np.typing.NDArray[np.float32]) -> float:
            nulls = np.isnan(data)
            if 0 in data.shape:
                log.warning(
                    "Validation region has 0 area, check input slices correspond"
                    "to coordinate values in the dataset",
                )
                return 1.0
            return float(nulls.sum() / np.prod(nulls.shape))

        if "latitude" in store_da.dims:
            spatial_dims: list[str] = ["latitude", "longitude"]
        elif "x_osgb" in store_da.dims:
            spatial_dims = ["x_osgb", "y_osgb"]
        elif "x_laea" in store_da.dims:
            spatial_dims = ["x_laea", "y_laea"]
        else:
            return Failure(
                ValueError(
                    "Store does not have expected spatial dimensions. "
                    "Expected: ['latitude', 'longitude'], ['x_osgb', 'y_osgb'], "
                    "['x_laea', 'y_laea']. Got: {store_da.dims}.",
                ),
            )

        result = xr.apply_ufunc(
            _calc_null_percentage,
            store_da,
            input_core_dims=[spatial_dims],
            vectorize=True,
            dask="parallelized",
        )

        failed_image_count: int = (result > nans_in_image_threshold).sum().values
        total_image_count: int = result.size
        failed_image_percentage: float = failed_image_count / total_image_count
        if failed_image_percentage > images_failing_nan_check_threshold:
            log.warning(
                f"Dataset failed validation. "
                f"{failed_image_percentage:.2%} of images have greater than "
                f"{int(nans_in_image_threshold * 100)}% null values"
                f"({failed_image_count}/{total_image_count})",
            )
            return Success(True)
        log.info(
            f"{failed_image_count}/{total_image_count} "
            f"({failed_image_percentage:.2%}) of images have greater than "
            f"{int(nans_in_image_threshold * 100)}% null values",
        )
        return Success(False)

    def validate_store(self) -> ResultE[None]:
        """Validate the store.

        This method checks the store for the presence of all expected parameters.

        Returns:
            A bool indicating the result of the validation.
        """
        log.debug(f"Validating store at '{self.path}'")
        store_da: xr.DataArray = xr.open_dataarray(self.path, engine="zarr")

        # Check for NaNs
        has_nans_result = self._has_nans(store_da=store_da)
        if isinstance(has_nans_result, Failure):
            return has_nans_result
        else:
            if has_nans_result.unwrap():
                return Failure(
                    ValueError(
                        "Store contains NaN values. "
                        "Check the source data for missing values and reprocess the data.",
                    ),
                )

        # TODO: Use consistency checks instead
        """
        # Consistency check on the coordinates of the store
        coords_result = NWPDimensionCoordinateMap.from_xarray(store_da)
        match coords_result:
            case Failure(e):
                return Failure(e)
            case Success(coords):
                if coords != self.coordinate_map:
                    return Failure(ValueError(
                        "Coordinate consistency check failed: "
                        "Store coordinates do not match expected coordinates. "
                        f"Expected: {self.coordinate_map}. Got: {coords}.",
                    ))

        # Validity check on the parameters of the store
        for param in self.coordinate_map.variable:
            scan_result: ResultE[ParameterScanResult] = self.scan_parameter_values(p=param)
            match scan_result:
                case Failure(e):
                    return Failure(e)
                case Success(scan):
                    log.debug(f"Scanned parameter {param.name}: {scan.__repr__()}")
                    if not scan.is_valid or scan.has_nulls:
                        return Failure(ValueError("Parameter validation failed."))
        """

        return Success(None)

    def delete_store(self) -> ResultE[None]:
        """Delete the store."""
        if self.path.startswith("s3://"):
            import s3fs

            try:
                fs = s3fs.S3FileSystem(
                    anon=False,
                    client_kwargs={
                        "region_name": os.getenv("AWS_REGION", "eu-west-1"),
                        "endpoint_url": os.getenv("AWS_ENDPOINT_URL", None),
                    },
                )
                fs.rm(self.path, recursive=True)
            except Exception as e:
                return Failure(
                    OSError(
                        f"Unable to delete S3 store at path '{self.path}'."
                        "Ensure AWS credentials are correct and discoverable by botocore. "
                        f"Error context: {e}",
                    ),
                )
        else:
            try:
                shutil.rmtree(self.path)
            except Exception as e:
                return Failure(
                    OSError(
                        f"Unable to delete store at path '{self.path}'. " f"Error context: {e}",
                    ),
                )
        log.info("Deleted zarr store at '%s'", self.path)
        return Success(None)

    def scan_parameter_values(self, p: Parameter) -> ResultE[ParameterScanResult]:
        """Scan the values of a parameter in the store.

        Extracts data from the values of the given parameter in the store.
        This reads the data from the store, so note that this can be an
        expensive operation for large datasets.

        Args:
            p: The name of the parameter to scan.

        Returns:
            A ParameterScanResult object.
        """
        if p not in self.coordinate_map.variable:
            return Failure(
                KeyError(
                    "Parameter scan failed: "
                    f"Cannot validate unknown parameter: {p.name}. "
                    "Ensure the parameter has been renamed to match the entities "
                    "parameters defined in `entities.parameters` if desired, or "
                    "add the parameter to the entities parameters if it is new. "
                    f"Store parameters: {[p.name for p in self.coordinate_map.variable]}.",
                ),
            )
        store_da: xr.DataArray = xr.open_dataarray(self.path, engine="zarr")

        # Calculating the mean of a dataarray returns another dataarray, so it
        # must be converted to a numpy array via `values`. Even though it is a
        # single number in this case, type checkers don't know that, so the
        # second call to `mean()` helps to reassure them its a float.
        mean = store_da.mean().values.mean()

        return Success(
            ParameterScanResult(
                mean=mean,
                is_valid=True,
                has_nulls=False,
            ),
        )

    def postprocess(self, options: PostProcessOptions) -> ResultE[str]:
        """Post-process the store.

        This creates a new store, as many of the postprocess options require
        modifications to the underlying file structure of the store.
        """
        # TODO: Implement postprocessing options
        if options.requires_postprocessing():
            log.info("Applying postprocessing options to store %s", self.name)

            if options.validate:
                log.warning("Validation not yet implemented in efficient manner. Skipping option.")

            log.debug("Postprocessing complete for store %s", self.name)
            return Success(self.path)

        else:
            return Success(self.path)

    def update_attrs(self, attrs: dict[str, str]) -> ResultE[str]:
        """Update the attributes of the store.

        This method updates the attributes of the store with the given dictionary.
        """
        group: zarr.Group = zarr.open_group(self.path)
        group.attrs.update(attrs)
        zarr.consolidate_metadata(self.path)
        return Success(self.path)

    def missing_times(self) -> ResultE[list[dt.datetime]]:
        """Find the missing init_time in the store.

        A "missing init_time" is determined by the values corresponding
        to the first two coordinate values of each dimension: if all are
        NaN or None values then the time is considered missing.
        """
        try:
            store_da: xr.DataArray = xr.open_dataarray(self.path, engine="zarr")
        except Exception as e:
            return Failure(
                OSError(
                    "Cannot determine missing times in store due to "
                    f"error reading '{self.path}': {e}",
                ),
            )
        missing_times: list[dt.datetime] = []
        for it in store_da.coords["init_time"].values:
            if (
                store_da.sel(init_time=it)
                .isel({d: slice(0, 2) for d in self.coordinate_map.dims if d != "init_time"})
                .isnull()
                .all()
                .values
            ):
                missing_times.append(pd.Timestamp(it).to_pydatetime().replace(tzinfo=dt.UTC))
        if len(missing_times) > 0:
            log.debug(
                f"NaNs in init times '{missing_times}' suggest they are missing, "
                f"will redownload",
            )
        return Success(missing_times)

    @staticmethod
    def _create_zarrstore_s3(s3_folder: str, filename: str) -> ResultE[tuple[MutableMapping, str]]:  # type: ignore
        """Create a mutable mapping to an S3 store.

        Authentication with S3 is done via botocore's credential discovery.

        Returns:
            A tuple containing the store mapping and the path to the store,
            in a result object indicating success or failure.

        See Also:
          - https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html#configuring-credentials
        """
        import s3fs

        if not s3_folder.startswith("s3://"):
            return Failure(
                ValueError(
                    "S3 folder path must start with 's3://'. " f"Got: {s3_folder}",
                ),
            )
        log.debug("Attempting AWS connection using credential discovery")
        try:
            fs = s3fs.S3FileSystem(
                anon=False,
                client_kwargs={
                    "region_name": os.getenv("AWS_REGION", "eu-west-1"),
                    "endpoint_url": os.getenv("AWS_ENDPOINT_URL", None),
                },
            )
            path = s3_folder + "/" + filename
            fs.mkdirs(path=path, exist_ok=True)
            store = s3fs.mapping.S3Map(path, fs, check=False, create=True)
        except Exception as e:
            return Failure(
                OSError(
                    f"Unable to create file mapping for path '{path}'. "
                    "Ensure ZARRDIR environment variable is specified correctly, "
                    "and AWS credentials are discoverable by botocore. "
                    f"Error context: {e}",
                ),
            )
        return Success((store, path))

    @staticmethod
    def gen_store_filename(coords: NWPDimensionCoordinateMap) -> str:
        """Create a filename for the store.

        If the store only covers a single init_time, the filename is the init time.
        Else, if it covers multiple init_times, the filename is the range of init times.
        The extension is '.zarr'.
        """
        store_range: str = coords.init_time[0].strftime("%Y%m%d%H")
        if len(coords.init_time) > 1:
            store_range = f"{coords.init_time[0]:%Y%m%d%H}-{coords.init_time[-1]:%Y%m%d%H}"

        return store_range + ".zarr"
